;; WB-tree File Based Associative String Data Base System.
;; Copyright (c) 1991, 1992, 1993, 2000 Free Software Foundation, Inc.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place, Suite 330, Boston,
;; MA 02111, USA.
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of GUILE.
;;
;; The exception is that, if you link the GUILE library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the GUILE library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name GUILE.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; GUILE, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for GUILE, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;; routines in this file return success codes

(require (in-vicinity (program-vicinity) "wbsys.scm"))

;;@node SEGs, HANDs and Tree Operations, Concepts, C Interface
;;@section SEGs
;;@cindex SEG

;; Segment not in use?
(define (seg:free? seg)
  (if (not lck-tab) (init-wb 75 150 2048))
  (cond ((or (negative? seg) (>= seg NUM-SEGS))
	 (edprintf "bad segment number %d\n" seg)
	 #f)
	((and (not (SEG:PORT seg))
	      (not (SEG:STR seg))
	      (not (SEG:USED seg)))
	 #t)
	(else #f)))

;TBD - need to lck seg here.
;; Segment will be read-only if MODE is #f.

;;@body 
;;@1 should be a non-negative number.  Opens the database @1 in @2 and
;;returns @1 if successful, a status code otherwise.  The database @1
;;will be read-only if the @3 argument is 0.  It will be read-write if
;;the @3 argument is 2.
(define (open-seg seg filename mode)
  (define bsiz #f)
  (define (errout reason-str)
    (edprintf "not a database %s %s\n" filename reason-str)
    (blk:file-close (SEG:PORT seg))
    (SEG:SET-PORT! seg #f)
    (SEG:SET-STR! seg #f)
    (SEG:SET-USED! seg #f)
    TYPERR)
  (if (zero? mode) (set! mode #f))
  (cond
   ((not (seg:free? seg))
    (edprintf "%s: segment in use %d\n" 'OPEN-SEG seg)
    ARGERR)
   ((begin
      (set! bsiz (min-file-blk-size filename))
      (set! bsiz (max (* 3 128) bsiz))
      ;;temporarily set bsiz so that we can get it from superblk
      (> bsiz blk-size))
    (edprintf "unsupported %s %d > %d\n" 'bsiz bsiz blk-size)
    ARGERR)
   (else
    (let loop ((file (if mode (blk:file-open-modify filename bsiz)
			 (blk:file-open-read-only filename bsiz))))
      (cond
       ((if mode (output-port? file) (input-port? file))
	(SEG:SET-PORT! seg file)
	(SEG:SET-STR! seg filename)
	(SEG:SET-USED! seg 2)
	(SEG:SET-BSIZ! seg bsiz)
	(SEG:SET-FLC-LEN! seg (if mode -1 -2)) ;-1 means to read in "FLC" image.
					;-2 means read only.
	(let ((han (SEG:RT-HAN seg))
	      (tmp-str (make-string 5))) ;this should be longer
	  (cond
	   ((err? (bt-open seg 0 han (+ WCB-SAP WCB-SAR))) ; superblock
	    (errout "bt-open 0"))
	   ((not (eq? 2 (bt-get han "BSIZ" 4 tmp-str)))
	    (errout "BSIZ"))
	   ((not (= bsiz (str2short tmp-str 0)))
	    (blk:file-close file)
	    (set! bsiz (str2short tmp-str 0))
	    (cond
	     ((> bsiz blk-size) (errout "BSIZ too big."))
	     (else (loop (if mode (blk:file-open-modify filename bsiz)
			     (blk:file-open-read-only filename bsiz))))))
	   ((not (eq? 4 (bt-get han "USED" 4 tmp-str)))
	    (errout "USED"))
	   (else
	    (SEG:SET-USED! seg (str2long tmp-str 0))
	    (cond ((not (eq? 5 (bt-get han "FLD" 3 tmp-str)))
		   (errout "FLD"))
		  ((err? (bt-open seg (str2long tmp-str 1)
				  (SEG:FL-HAN seg) WCB-SAR))
		   (errout "FLC"))
		  (else
		   (if (not  (eqv? (HAN-TYP (SEG:FL-HAN seg)) FRL-TYP))
		       (dprintf "Older type freelist - still supported.\n"))
		   (HAN-SET-WCB! (SEG:FL-HAN seg) WCB-SAR)
		   seg))))))
       (else
	(if (if mode (input-port? file) (output-port? file)) (blk:file-close file))
	(edprintf "could not open file %s\n" filename)
	IOERR))))))

;;@body
;;Closes database segment @1 and the file containing it.  If @2 is
;;NULL, then if there are any problems freeing buffers, then the close
;;is aborted.  A status code is returned.
(define (close-seg seg hammer)
  (cond ((or (not (SEG:STR seg))
	     (not (SEG:USED seg)))
	 (edprintf "%s: segment %d already closed\n" 'CLOSE-SEG seg)
	 ARGERR)
	(else
	 (flush-flc! seg 5)
	 "leave only enough blocks to fit in flc in superblock."
	 (if (>= (SEG:FLC-LEN seg) 0)
	     (let* ((tmp-str (make-string 20)))
	       (do ((i (+ -1 (SEG:FLC-LEN seg)) (+ -1 i)))
		   ((negative? i))
		 (long2str! tmp-str (* 4 i) (vector-ref (SEG:FLC seg) i)))
	       (bt-put (SEG:RT-HAN seg) "FLC" 3 tmp-str (* 4 (SEG:FLC-LEN seg)))
	       (SEG:SET-FLC-LEN! seg -1)))
	 (let ((ans (do-seg-buffers seg flush-buffer)))
	   (cond ((or (success? ans) hammer)
		  (if (not (success? ans)) (set! ans NOTPRES))
		  (do-seg-buffers seg purge-buffer)
		  (bt-close (SEG:RT-HAN seg))
		  (bt-close (SEG:FL-HAN seg))
		  (blk:file-close (SEG:PORT seg))
		  (SEG:SET-PORT! seg #f)
		  (SEG:SET-STR! seg #f)
		  (SEG:SET-USED! seg #f)))
	   ans))))

(define db-version-str "wb1a6")
(define db-authors-str "A. Jaffer, J. Finger, R. Zito-Wolf")

;;@body
;;
;;The integer @1 must specify an unused segment.  The integer @3
;;specifies the size of B-tree blocks.  Unless you are testing
;;splitting algorithms, @3 should be 2048.
;;
;;@0 makes a new empty database named @2.  A status code is returned.
;;Call @cname{open-seg} to use the new database.
(define (make-seg seg filename bsiz)
  (cond
   ((or (not (seg:free? seg)) (not (try-lck (SEG:LCK seg))))
    (edprintf "%s: segment in use %d\n" 'MAKE-SEG seg)
    ARGERR)
   ((> bsiz blk-size)
    (edprintf "unsupported %s %d > %d\n" 'bsiz bsiz blk-size)
    (unlck! (SEG:LCK seg))
    ARGERR)
   (else
    (let ((file (blk:file-create filename bsiz)))
      (cond
       ((output-port? file)
	(SEG:SET-PORT! seg file)
	(SEG:SET-BSIZ! seg bsiz)
	(SEG:SET-USED! seg 3)
	(SEG:SET-STR! seg filename)
	(init-leaf-blk! empty-blk 0 DIR-TYP)
	(BLK:SET-TIME! empty-blk (current-time))
	(blk:write file empty-blk bsiz 0)
	(init-leaf-blk! empty-blk 1 DIR-TYP)
	(BLK:SET-TIME! empty-blk (current-time))
	(blk:write file empty-blk bsiz 1)
	(init-leaf-blk! empty-blk 2 FRL-TYP)
	(BLK:SET-TIME! empty-blk (current-time))
	(blk:write file empty-blk bsiz 2)
	(blk:file-close file)
	(set! file (blk:file-open-modify filename bsiz))
	(cond ((output-port? file)
	       (SEG:SET-PORT! seg file)
	       (unlck! (SEG:LCK seg))
	       (let ((han (SEG:RT-HAN seg))
		     (tmp-str (make-string 5)))
		 (bt-open seg 0 han (+ WCB-SAP WCB-SAR))
		 (bt-put han "" 0
			 db-version-str (string-length db-version-str))
		 (long2str! tmp-str 0 (SEG:USED seg))
		 (bt-put han "USED" 4 tmp-str 4)
		 (short2str! tmp-str 0 (SEG:BSIZ seg))
		 (bt-put han "BSIZ" 4 tmp-str 2)
		 (string-set! tmp-str 0 (integer->char 4))
		 (long2str! tmp-str 1 1)
		 (bt-put han "ROOT" 4 tmp-str 5)
		 (long2str! tmp-str 1 2)
		 (bt-put han "FLD" 3 tmp-str 5)
		 (bt-put han "FLC" 3 "" 0)
		 (if (> bsiz 128)
		     (bt-put han "authors" 7
			     db-authors-str (string-length db-authors-str)))
		 (close-seg seg #f)	;don't close the segment if it is memory resident.
		 ))
	      (else (edprintf "couldn't open fresh file %s\n" filename)
	       (unlck! (SEG:LCK seg))
	       IOERR)))
       (else (edprintf "couldn't create new file %s\n" filename)
	     (unlck! (SEG:LCK seg))
	     IOERR))))))

;;@cindex wcb
;;@noindent The write-control-bits argument (@var{wcb}) to these
;;functions controls the latency of updates to the file after various
;;operations.  These bits are defined as follows:
;;  
;;@multitable @columnfractions .1 .3 .6
;;@item @var{value}
;;@tab C-name
;;@tab Meaning
;;@item 1
;;@tab wcb_sap
;;@tab save block after PUTs
;;@cindex wcb_sap
;;@item 2
;;@tab wcb_sar
;;@tab save block after REMOVEs
;;@cindex wcb_sar
;;@item 4
;;@tab wcb_sac
;;@cindex wcb_sac
;;@tab force block save after cached block changes
;;@item 8
;;@tab wcb_fac
;;@tab flush buffer entirely after cached block changes (not currently
;;implemented)
;;@cindex wcb_fac
;;@end multitable

;;@body Opens bt-handle @3 to seg number @1, block number @2, and
;;returns the type of the block.  If no such block exists or is not a
;;root block, then a (negative) status code is returned.
(define (bt-open seg blk-num han wcb)
  (if (and (>= seg 0) (< seg NUM-SEGS) (SEG:STR seg)) ;allocated
      (let ((ent (get-ent seg blk-num ACCREAD)))
	(cond ((not ent) UNKERR)
	      ((not (root? (ENT:BLK ent)))
	       (release-ent! ent ACCREAD)
	       (edprintf "%s: not a root %d:%ld\n" 'BT-OPEN seg blk-num)
;;;	       (check-access!)
	       ARGERR)
	      (else
	       (HAN-SET-SEG! han seg)
	       (HAN-SET-NUM! han blk-num)
	       (HAN-SET-TYP! han (BLK:TYP (ENT:BLK ent))) ;TBD improve. (eh?)
	       (HAN-SET-LAST! han blk-num)
	       (if (BLK:TYP? (ENT:BLK ent) DIR-TYP)
		   (set! wcb (logior wcb (+ WCB-SAP WCB-SAR))))
	       (HAN-SET-WCB! han wcb)
	       (release-ent! ent ACCREAD)
;;;	       (check-access!)
	       (HAN-TYP han))))
      ARGERR))

;;@body Creates a new root block in seg @1 of type @2, opens bt-handle
;;@3 to it, and returns a status code.  If @1 has insufficient room to
;;create a new tree, then the @var{noroom} status code is returned.
;;
;;@0 can be used to create temporary b-trees.  Temporary trees will be
;;be reclaimed by check program after system crashes.  In order to
;;make a tree persistent, add it to a directory (tree).
(define (bt-create seg typ han wcb)
  (define ent (create-new-blk-ent seg))
  (cond ((not ent) NOROOM)
	(else (let* ((blk:num (ENT:ID ent)))
	     (init-leaf-blk! (ENT:BLK ent) blk-num typ)
	     (ENT:SET-DTY! ent #t)
	     (ENT:SET-PUS! ent 0)
	     (ent:write ent)
	     (HAN-SET-SEG! han seg)
	     (HAN-SET-NUM! han blk-num)
	     (HAN-SET-TYP! han typ)
	     (HAN-SET-LAST! han blk-num)
	     (if (eqv? typ DIR-TYP)
		 (set! wcb (logior wcb (+ WCB-SAP WCB-SAR))))
	     (HAN-SET-WCB! han wcb)
	     (release-ent! ent ACCWRITE)
;;;	     (check-access!)
	     SUCCESS))))


;;@body Closes bt-handle @1 and returns @var{SUCCESS}.
;;
;;Currently, @0 has no effect other than to clear @1.
(define (bt-close han)
  (HAN-SET-SEG! han 0)
  (HAN-SET-NUM! han 0)
  (HAN-SET-TYP! han 0)
  (HAN-SET-LAST! han 0)
  SUCCESS)

;;@node HANDs and Tree Operations, Scan, SEGs, C Interface
;;@section HANDs and Tree Operations
;;@cindex HAND

(define clever-cache-enable #t)

;;@noindent
;;Note: most of the data-manipulating commands here
;;can return @var{notpres}, with the followng meanings:
;;@cindex notpres
;;
;;@multitable @columnfractions .2 .8
;;@item @code{bt-get}
;;@tab no such key.
;;@item @code{bt-next}
;;@tab no NEXT key (ie, key given was LAST key).
;;@item @code{bt-prev}
;;@tab no PREV key (ie, key given was FIRST key).
;;@item @code{bt-rem}
;;@tab KEY was not found.
;;@c @item @code{bt-rem-range}
;;@c @tab ??
;;@item @code{bt-put}
;;@tab NOT USED (could be symmetric w/WRITE).
;;@item @code{bt-write}
;;@tab key WAS found, so no write done.
;;@end multitable

;;@body @2 is a string of length @3.  @0 stores into the string @4 the
;;value associated with @2 in tree @1.  @0 returns the length of the
;;string stored into @4 or an error code.
(define (bt-get han key-str k-len ans-str)
  (define pkt (make-vector PKT-SIZE))
  (define ent #f)
;;;  (dprintf "%s %d:%ld %.*s\n" 'BT-GET (HAN-SEG han) (HAN-ID han)
;;;	   (max 0 k-len) key-str)
  (set! ent (chain-find-ent han ACCREAD key-str k-len pkt))
  (cond ((not ent) (set! get-fct (+ 1 get-fct))
		    UNKERR)
	((not (eq? (MATCH-TYPE pkt) MATCH))
	 (set! get-ct (+ 1 get-ct))
	 (release-ent! ent ACCREAD)
	 NOTPRES)
	(else
	 (let ((alen (get-this-val (ENT:BLK ent) (MATCH-POS pkt) ans-str)))
	   (set! get-ct (+ 1 get-ct))
	   (release-ent! ent ACCREAD)
	   alen))))

;;@body @2 is a string of length @3.  @0 stores into the string @4 the
;;next key after @2 in tree @1.  @0 returns the length of the string
;;stored into @4 or an error code.
(define (bt-next han key-str k-len ans-str)
  (define pkt (make-vector PKT-SIZE))
  (define ent #f)
;;;(dprintf "%s %d:%ld %.*s\n" 'BT-NEXT (HAN-SEG han) (HAN-ID han)
;;;	   (max 0 k-len) key-str)
  (set! ent (chain-find-ent han ACCREAD key-str k-len pkt))
  (cond ((not ent)
	 (set! next-fct (+ 1 next-fct))
	 UNKERR)
	(else
	 (set! next-ct (+ 1 next-ct))
	 (let ((res (chain-next ent key-str k-len ans-str pkt)))
	   (if clever-cache-enable (HAN-SET-LAST! han (BLK:TO-CACHE pkt)))
	   res))))

;;@body @2 is a string of length @3.  @0 stores into the string @4 the
;;last key before @2 in tree @1.  @0 returns the length of the string
;;stored into @4 or an error code.
(define (bt-prev han key-str k-len ans-str)
  (define pkt (make-vector PKT-SIZE))
  (define ent #f)
;;;(dprintf "%s %d:%ld %.*s\n" 'BT-PREV (HAN-SEG han) (HAN-ID han)
;;;	   (max 0 k-len) key-str)
  (set! ent (chain-find-prev-ent han ACCREAD key-str k-len pkt))
  (and ent (set! ent (prev-k-ent ent key-str k-len LEAF pkt)))
  (cond ((not ent)
	 (set! prev-fct (+ 1 prev-fct))
	 UNKERR)
	(else
	 (set! prev-ct (+ 1 prev-ct))
	 (if (zero? (MATCH-POS pkt))
	     (begin (release-ent! ent ACCREAD) NOTPRES)
	     (let ((k-len2 (recon-this-key (ENT:BLK ent)
					   (MATCH-POS pkt) ans-str 0 256)))
	       (HAN-SET-LAST! han (ENT:ID ent))
	       (release-ent! ent ACCREAD)
	       k-len2)))))


;;@body @2 is a string of length @3.  @0 stores into the string @4 the
;;value associated with @2 in tree @1; then removes that association
;;from tree @1.  @0 returns the length of the string stored into @4 or
;;an error code.
;;
;;If @4 is 0, @0 removes the @2 association from tree @1 and returns
;;@var{SUCCESS} if successful; an error code if not.
(define (bt-rem han key-str k-len ans-str)
  (define pkt (make-vector PKT-SIZE))
  (define ent #f)
;;;(dprintf "%s %d:%ld %.*s\n" 'BT-REM (HAN-SEG han) (HAN-ID han)
;;;	   (max 0 k-len) key-str)
  (cond ((< k-len 0)
	 (edprintf "%s: bad length string %d\n" 'BT-REM k-len)
	 ARGERR)
	(else
	 (set! ent (chain-find-ent han ACCWRITE key-str k-len pkt))
	 (cond (ent
		(set! rem-ct (+ 1 rem-ct))
		(let ((ans (chain-rem ent key-str k-len ans-str pkt (HAN-WCB han))))
		  (release-ent! ent ACCWRITE)
		  ans))
	       (else
		(set! rem-fct (+ 1 rem-fct))
		UNKERR)))))

;;; To make possible bounded-time operation, @0 will purge at most
;;; BLK-LIMIT blocks at a time; passing BLK-LIMIT of -1 imposes no
;;; limit.

;;@body @2 must be a maximum-length (256 byte) string containing a key
;;@3 bytes long.  @4 is a string of length @5.
;;
;;@0 removes [@2 @dots{} @4) and their values.  If @4 <= @2 no
;;deletion will occur (even if @2 is found).  @0 returns SUCCESS if
;;the operation is complete, an error status code if not.
(define (bt-rem-range han key-str k-len key2-str k2-len)
  (define respkt (make-vector PKT-SIZE))
  (bt-scan han REM-SCAN key-str k-len key2-str k2-len #f #f respkt -1))

;;@body @2 is a string of length @3.  @4 is a string of length @5.  @0
;;makes the value associated with @2 be @4 in tree @1.  @0 returns a
;;status code for the operation.
(define (bt-put han key-str k-len val-str v-len)
  (define ent #f)
  (define pkt (make-vector PKT-SIZE))
;;;(dprintf "%s %d:%ld %.*s %.*s\n" 'BT-PUT
;;;	   (HAN-SEG han) (HAN-ID han) (max 0 k-len) key-str v-len val-str)
  (cond ((or (> v-len 255) (> k-len 255) (< k-len 0))
	 ARGERR)
	(else
	 (set! ent (chain-find-ent han ACCWRITE key-str k-len pkt))
	 (if ent
	     (let ((res (chain-put ent key-str k-len val-str v-len pkt #f (HAN-WCB han))))
	       (cond (res
		      (if clever-cache-enable
			  (HAN-SET-LAST! han (BLK:TO-CACHE pkt)))
		      (set! put-ct (+ 1 put-ct))
		      SUCCESS)
		     (else
		      (set! put-fct (+ 1 put-fct))
		      UNKERR)))
	     UNKERR))))

;;@body @2 is a string of length @3.  @4 is a string of length @5.  If
;;@1 currently contains an association for @2, then @0 does not modify
;;the tree and returns the @var{notpres} status code.
;;
;;Otherwise, @0 makes the value associated with @2 be @4 in tree @1.
;;@0 returns a status code for the operation.
(define (bt-write han key-str k-len val-str v-len)
  (define ent #f)
  (define pkt (make-vector PKT-SIZE))
  (cond
   ((or (> v-len 255) (> k-len 255) (< k-len 0))
    ARGERR)
   (else
    (set! ent (chain-find-ent han ACCWRITE key-str k-len pkt))
    (if ent
	(if (eq? (MATCH-TYPE pkt) MATCH)
	    (begin (release-ent! ent ACCWRITE) NOTPRES) ;DTY has not been set.
	    (let ((res (chain-put ent key-str k-len val-str v-len pkt #f (HAN-WCB han))))
	      (cond (res
		     (if clever-cache-enable
			 (HAN-SET-LAST! han (BLK:TO-CACHE pkt)))
		     (set! put-ct (+ 1 put-ct))
		     SUCCESS)
		    (else
		     (set! put-fct (+ 1 put-fct))
		     UNKERR))))
	UNKERR))))
